/* eslint-disable no-console */
import type { ThemedToken, ThemeRegistration } from '@shikijs/types'
import type { ImageStyle, ScrollViewProps, StyleProp, TextStyle, ViewStyle } from 'react-native'
import type { SyntaxHighlighterProps } from 'src/types/shiki'
import transform, { type StyleTuple } from 'css-to-react-native'
import React, { useMemo } from 'react'
import { Platform, ScrollView, StyleSheet, Text, View } from 'react-native'
import { useSyntaxHighlighter } from '../../hooks/useSyntaxHighlighter'

const monospaceFont = Platform.select({
  ios: 'Menlo',
  android: 'monospace',
  default: 'monospace',
})

const DEBUG = true

type NamedStyles<T> = { [P in keyof T]: ViewStyle | TextStyle | ImageStyle }

function getThemeStyles(theme: ThemeRegistration) {
  const styles: Record<string, TextStyle> = {}

  theme.tokenColors?.forEach((token) => {
    if (!token.settings || !token.scope)
      return

    const scopes = Array.isArray(token.scope) ? token.scope : typeof token.scope === 'string' ? token.scope.split(',').map((scope: string) => scope.trim()) : []

    const style = cleanStyle({
      color: token.settings.foreground,
      fontStyle: token.settings.fontStyle,
      fontWeight: token.settings.fontStyle === 'bold' ? 'bold' : 'normal',
    })

    scopes.forEach((scope: string) => {
      styles[scope] = style
    })
  })

  DEBUG && console.log('Theme styles:', styles)
  return styles
}

function cleanStyle(style: Record<string, any>): TextStyle {
  const filtered = Object.entries(style)
    .filter(([_key, value]) => value != null)
    .map<StyleTuple>(([key, value]) => [key, value])

  DEBUG && console.log('Cleaning style:', style, 'Filtered:', filtered)
  return transform(filtered) as TextStyle
}

interface Props extends SyntaxHighlighterProps {
  scrollViewProps?: ScrollViewProps
  textStyle?: StyleProp<TextStyle>
  containerStyle?: StyleProp<ViewStyle>
}

const SyntaxHighlighter: React.FC<Props> = ({ text, language, languageId, theme, fontSize = 14, scrollViewProps, textStyle, containerStyle }) => {
  const { tokens } = useSyntaxHighlighter({
    text,
    language,
    languageId: languageId || String(language),
    theme,
  })

  const themeStyles = useMemo(() => getThemeStyles(theme as ThemeRegistration), [theme])

  const baseStyles = useMemo(() => {
    const styles: NamedStyles<{
      scrollView: ViewStyle
      codeContainer: ViewStyle
      lineContainer: ViewStyle
      token: TextStyle
    }> = {
      scrollView: {
        flex: 1,
        minHeight: 20,
      },
      codeContainer: {
        padding: 16,
        minWidth: '100%' as const,
        backgroundColor: theme.colors?.['editor.background'] || '#1e1e1e',
      },
      lineContainer: {
        flexDirection: 'row',
        flexWrap: 'wrap',
      },
      token: {
        fontFamily: monospaceFont,
        fontSize: 14,
        color: theme.colors?.['editor.foreground'] || '#d4d4d4',
      },
    }
    return StyleSheet.create(styles)
  }, [theme])

  const renderToken = (token: ThemedToken, rowIndex: number, tokenIndex: number) => {
    const defaultStyle = baseStyles.token
    const scopeStyles = token.explanation?.flatMap(exp => exp.scopes.map(scope => themeStyles[scope.toString()]).filter(Boolean)) || []
    const specificStyle = token.color
      ? cleanStyle({
        color: token.color,
        fontStyle: token.fontStyle ? 'italic' : 'normal',
        fontWeight: token.fontStyle ? 'bold' : 'normal',
      })
      : undefined

    DEBUG && console.log('Token:', token, 'Scope styles:', scopeStyles)

    return (
      <Text key={`${rowIndex}-${tokenIndex}`} style={[defaultStyle, ...scopeStyles, specificStyle, textStyle, { fontSize }]}>
        {token.content}
      </Text>
    )
  }

  return (
    <ScrollView
      horizontal
      {...scrollViewProps}
      style={[baseStyles.scrollView, scrollViewProps?.style]}
      contentContainerStyle={[baseStyles.codeContainer, scrollViewProps?.contentContainerStyle, containerStyle]}
    >
      <View onStartShouldSetResponder={() => true}>
        {tokens.map((line, rowIndex) => (
          <View key={rowIndex} style={baseStyles.lineContainer}>
            {line.map((token, tokenIndex) => renderToken(token, rowIndex, tokenIndex))}
          </View>
        ))}
      </View>
    </ScrollView>
  )
}

export default SyntaxHighlighter
